--[[
    ActionBarLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    ActionBarLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ActionBarLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ActionBarLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
  The ActionBarRenderer class.
--]]

local actionBarLib = LibStub:NewLibrary("ActionBarLib-1.0", 1);
if not actionBarLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
-- end import

local ActionBarRenderer, ActionBarRendererFactory = Factory.CreateAbstractType();
actionBarLib.ActionBarRenderer = ActionBarRenderer;
actionBarLib.ActionBarRendererFactory = ActionBarRendererFactory;

ActionBarRenderer.TOOLTIP_HIDE_IN_COMBAT = 1;
ActionBarRenderer.TOOLTIP_SHOW = 2;
ActionBarRenderer.TOOLTIP_HIDE = 3;

ActionBarRenderer.tooltipVisibility = 1;
ActionBarRenderer.tooltipHelpAdded = true;

function ActionBarRenderer:ApplySkin(secureUnitButton)
  error("ActionBarRenderer:ApplySkin must be overrided");
end

function ActionBarRenderer:UnApplySkin(secureUnitButton)
  error("ActionBarRenderer:UnApplySkin must be overrided");
end

function ActionBarRenderer:ActionChanged(secureUnitButton)
  error("ActionBarRenderer:ActionChanged must be overrided");
end

function ActionBarRenderer:IsTooltipAsked()
  if self.tooltipVisibility == 2 then
    return true;
  elseif self.tooltipVisibility == 3 then
    return false;
  else
    return not InCombatLockdown();
  end
end

function ActionBarRenderer:AddSelectedTooltipHelp(tooltip)
  tooltip:AddLine("Action bar selected button (bindable)", 0, 1, 0);
  tooltip:AddLine("Alt-leftClick : Add a favorite", 1, 1, 1);
  tooltip:AddLine("Shift-alt-leftClick : Remove a favorite", 1, 1, 1);
  tooltip:AddLine("RightClick : Open the bar", 1, 1, 1);
  tooltip:AddLine("LeftClick : Perform action", 1, 1, 1);
 -- tooltip:AddLine("Shift-leftClick : Perform action first rank", 1, 1, 1);
  tooltip:AddLine("Ctrl-leftClick : May cancel action result", 1, 1, 1);
end

function ActionBarRenderer:AddFavoriteTooltipHelp(tooltip)
  tooltip:AddLine("Action bar favorite button", 0, 1, 0);
  tooltip:AddLine("Ctrl-leftclick : Exchange action with next", 1, 1, 1);
  tooltip:AddLine("Shift-ctrl-leftclick : Exchange action with previous", 1, 1, 1);
  tooltip:AddLine("Rightclick : Exchange action with selected action", 1, 1, 1);
  tooltip:AddLine("Leftclick : Perform action", 1, 1, 1);
 -- tooltip:AddLine("Shift-leftclick : Perform action first rank", 1, 1, 1);
end

function ActionBarRenderer:AddOpenTooltipHelp(tooltip)
  tooltip:AddLine("Action bar button", 0, 1, 0);
  tooltip:AddLine("Ctrl-leftclick : Exchange action with next", 1, 1, 1);
  tooltip:AddLine("Shift-ctrl-leftclick : Exchange action with previous", 1, 1, 1);
  tooltip:AddLine("Rightclick : Close the bar and select action", 1, 1, 1);
  tooltip:AddLine("Leftclick : Close the bar, select action and perform action", 1, 1, 1);
 -- tooltip:AddLine("Shift-leftclick : Close the bar, select action and perform action first rank", 1, 1, 1);
end


--[[
  The ActionBarState class.
--]]

local ActionBarState, ActionBarStateFactory, actionBarStateFactory = Factory.CreateType();

function ActionBarState:IsOpened()
  return self.selectedVisibility == 0;
end

-- 0,1 open ; 2,3 close ; 1,3 stuck
function ActionBarState:AddFavorite()
  local actionListSize = #self.visibilityList;
  if self.favoriteCount<actionListSize then
    self.favoriteCount = self.favoriteCount + 1;
    local oldVisibility = self.visibilityList[self.favoriteCount];
    if oldVisibility==0 then
      self.visibilityList[self.favoriteCount] = 1;
    elseif oldVisibility==2 then
      self.visibilityList[self.favoriteCount] = 3;      
    end
  end
end

function ActionBarState:RemoveFavorite()
  if self.favoriteCount>0 then
    local oldVisibility = self.visibilityList[self.favoriteCount];
    if oldVisibility==1 then
      self.visibilityList[self.favoriteCount] = 0;
    elseif oldVisibility==3 then
      self.visibilityList[self.favoriteCount] = 2;      
    end
    self.favoriteCount = self.favoriteCount - 1;
  end
end

function ActionBarState:Open()
  if self.selectedVisibility == 0 then
    error("ActionBarState:Open already open");
  end
  self.selectedVisibility = 0;
  self.selectedAction = 0;
  self.selectedPosition = 0;
  local actionListSize = #self.visibilityList;
  for i=1, actionListSize do
    if self.visibilityList[i] == 4 then
      break;
    elseif self.visibilityList[i] == 2 then
      self.visibilityList[i] = 0;
    elseif self.visibilityList[i] == 3 then
      self.visibilityList[i] = 1;
    end
  end
  self.originalSelectedAction = 0;
end

function ActionBarState:Swap(position)
  if self.selectedVisibility == 0 then
    error("ActionBarState:Swap cannot swap while open");
  end  
  self.selectedAction = self.actionList[position];
  self.originalSelectedAction = selectedAction;
  self.actionList[position] = self.actionList[self.selectedPosition];
  self.actionList[self.selectedPosition] = self.selectedAction;
end

function ActionBarState:Close(position)
  if self.selectedVisibility == 1 then
    error("ActionBarState:Close already closed");
  end
  self.selectedPosition = position;
  local actionId = self.actionList[position];
  self.selectedVisibility = 1;
  self.selectedAction = actionId;
  local actionListSize = #self.visibilityList;
  for i=1, actionListSize do
    if self.visibilityList[i] == 4 then
      break;
    elseif self.visibilityList[i] == 0 then
      self.visibilityList[i] = 2;    
    elseif self.visibilityList[i] == 1 then
      self.visibilityList[i] = 3;
    end
  end
  self.originalSelectedAction = actionId;
end

-- do not call directly
function ActionBarState:InsertAction(actionId, position, stucked)
  local actionListSize = #self.actionList;
  local positionOldVisibility = self.visibilityList[position];

  if positionOldVisibility==4 then
    -- insert at the end (imply no need to push button at the end)
    self.actionList[position] = actionId;
    if stucked then
      -- if previous is stuck
      if position == 1 or self.visibilityList[position - 1]%2 == 1 then
        self.favoriteCount = self.favoriteCount + 1;
        if self.selectedVisibility==0 then
          self.visibilityList[position] = 1;
        else
          self.visibilityList[position] = 3;        
        end
      else
        -- we can't be stuck so we becomes like previous
        self.visibilityList[position] = self.visibilityList[position - 1];      
      end
    else
      -- no issue we set our value
      if self.selectedVisibility==0 then
        self.visibilityList[position] = 0;
      else
        self.visibilityList[position] = 2;        
      end
    end
    return;
  end  
  
  for i=1, actionListSize - position do
    self.actionList[actionListSize-i+1] = self.actionList[actionListSize-i];
  end
  self.actionList[position] = actionId;
  local stateReplaced = false;
  
  local positionNewVisibility = positionOldVisibility;

  if stucked then    
    if position == 1 or self.visibilityList[position - 1]%2 == 1 then
      -- if the previous is stucked then we can set our value
      self.favoriteCount = self.favoriteCount + 1;
      if positionOldVisibility == 0 then
        self.visibilityList[position] = 1;
        positionNewVisibility = 1;
      elseif positionOldVisibility == 2 then
        self.visibilityList[position] = 3; 
        positionNewVisibility = 3;     
      end
    end
  end  
  
  local j = position + 1;
  -- first different
  while self.visibilityList[j] == positionNewVisibility do
    j = j + 1;
  end
  local secondPositionVisibility = self.visibilityList[j];

  -- if we get a non empty button
  if secondPositionVisibility~=4 then
    -- look for '4'
    local k = j + 1;
    while self.visibilityList[k] == secondPositionVisibility do
      k = k + 1;
    end
    self.visibilityList[k] = secondPositionVisibility;
    -- if we did not change our state
    if positionNewVisibility == positionOldVisibility then
      self.visibilityList[j] = positionOldVisibility;
    end
  else
    -- we find 4
    self.visibilityList[j] = positionOldVisibility;
  end
end

function ActionBarState:DeleteAction(position)
  local actionListSize = #self.actionList;
  if self.visibilityList[position]==1 or self.visibilityList[position]==3 then
    self.favoriteCount = self.favoriteCount - 1;
  end
  for i=position, actionListSize-1 do
    self.actionList[i] = self.actionList[i+1];
    self.visibilityList[i] = self.visibilityList[i+1];
  end
  self.actionList[actionListSize] = 0;
  self.visibilityList[actionListSize] = 4;  
end

function ActionBarState:EnableAction(actionId)
  local unavailable = self.unavailable[actionId];
  if unavailable==nil then
    error("ActionBarState:EnableAction already enabled");    
  end
  self.totalCount = self.totalCount + 1;
  local pos = unavailable[1];
  local i = pos;
  for _, v in pairs(self.unavailable) do
    if v[1] < unavailable[1] then
      i = i - 1;
    end
  end
  if self.selectedPosition >= i then
    self.selectedPosition = self.selectedPosition + 1;
  end
  self:InsertAction(actionId, i, unavailable[2]);
  self.unavailable[actionId] = nil;
  if self.originalSelectedAction == actionId then
    self:Close(i);
  end
end

function ActionBarState:IsActionEnabled(actionId)
  return (self.unavailable[actionId] == nil);
end

function ActionBarState:SetActionEnabled(actionId, enabled)
  if enabled then
    self:EnableAction(actionId);
  else
    self:DisableAction(actionId);
  end
end

function ActionBarState:DisableAction(actionId)
  local i = 1;
  local actionListSize = #self.actionList;
  while i<=actionListSize and self.actionList[i] ~= actionId do
    i = i + 1;
  end
  if i>actionListSize then
    error("ActionBarState:DisableAction already disabled or inexistant");  
  end
  self.totalCount = self.totalCount - 1;
  local defined = {};
  for _, v in pairs(self.unavailable) do
    defined[v[1]] = true;
  end

  local pos = 0;
  local j = 0; -- j = # button found
  while j<i do
    pos = pos + 1;
    if not defined[pos] then
      j = j + 1; -- there is a real button
    end
  end
  
  local stucked = (self.visibilityList[i]==1 or self.visibilityList[i]==3);
  
  self.unavailable[actionId] = {pos, stucked};
  if self.selectedPosition > i then
    -- if equals then menu will be open so selectedPosition will be lost
    -- that why no decrease with equals (save one operation)
    self.selectedPosition = self.selectedPosition - 1;
  end

  self:DeleteAction(i);
  
  if self.selectedAction == actionId then
    self:Open();
    self.originalSelectedAction = actionId;
  end
end

function ActionBarStateFactory:Create(actionListSize)
  local actionBarState = self:NewInstance();
  actionBarState.selectedVisibility = 0;
  actionBarState.selectedAction = 0;
  actionBarState.selectedPosition = 0;
  actionBarState.originalSelectedAction = 0;
  actionBarState.favoriteCount = 0;
  -- all are enabled
  actionBarState.totalCount = actionListSize;
  actionBarState.visibilityList = {};
  actionBarState.actionList = {};
  actionBarState.unavailable = {};

  for i=1,actionListSize do
    actionBarState.visibilityList[i] = 0;
    actionBarState.actionList[i] = i;
  end
  
  return actionBarState;
end

--[[
  The ActionBar class.
--]]


local ActionBar, ActionBarFactory, actionBarFactory = Factory.CreateType();
actionBarLib.ActionBar = ActionBar;
actionBarLib.ActionBarFactory = ActionBarFactory;
actionBarLib.actionBarFactory = actionBarFactory;

function ActionBar:SpreadState()

  local actionBarState = self.state;
  local opened = actionBarState:IsOpened();

  if self.state.selectedAction ~= 0 then
    local selectedAction = self.actionList[self.state.selectedAction];
    self.selectedSecureUnitButton:SetAttribute("type", selectedAction:GetType());
    self.selectedSecureUnitButton:SetAttribute(selectedAction:GetType(), selectedAction:GetValue());
    self.selectedSecureUnitButton:SetAttribute("unit", self.state.selectedAction);
    local target = selectedAction:GetTarget();
    if target == "target" then
      self.selectedSecureUnitButton:SetAttribute("unit", nil);
    else
      self.selectedSecureUnitButton:SetAttribute("unit", target);
    end
    self.selectedSecureUnitButton.action = selectedAction;
    self.actionBarRenderer:ActionChanged(self.selectedSecureUnitButton);
    self.selectedSecureUnitButton:Show();
  else
    self.selectedSecureUnitButton.action = nil;
    self.actionBarRenderer:ActionChanged(self.selectedSecureUnitButton);
    self.selectedSecureUnitButton:Hide();
  end

  for i=1, self.state.totalCount do    
    local currentAction = self.actionList[self.state.actionList[i]];
    self.secureUnitButtonList[i]:SetAttribute("type", currentAction:GetType())
    self.secureUnitButtonList[i]:SetAttribute(currentAction:GetType(), currentAction:GetValue())
    local target = currentAction:GetTarget();
    if target == "target" then
        self.secureUnitButtonList[i]:SetAttribute("unit", nil);
    else
        self.secureUnitButtonList[i]:SetAttribute("unit", target);
    end

    self.secureUnitButtonList[i].action = currentAction;
    self.actionBarRenderer:ActionChanged(self.secureUnitButtonList[i]);
    
    if opened then
      self.secureUnitButtonList[i]:Show();
    else
      if i<=self.state.favoriteCount then
        self.secureUnitButtonList[i]:Show();
      else
        self.secureUnitButtonList[i]:Hide();      
      end
    end        
  end
  for i=self.state.totalCount+1, #self.state.visibilityList do
    self.secureUnitButtonList[i].action = nil;
    self.actionBarRenderer:ActionChanged(self.secureUnitButtonList[i]);
    self.secureUnitButtonList[i]:Hide();
  end

  local openString = "true";
  if not actionBarState:IsOpened() then
    openString = "false";
  end
  self.header:Execute("open = "..openString.."; selectedPosition = "..actionBarState.selectedPosition.."; favoriteCount = "..actionBarState.favoriteCount.."; totalCount = "..actionBarState.totalCount..";");
    
end

-- disable an action (if selected and close we open)
function ActionBar:SetSelectedCtrlClickFunc(removeFunc)
  self.selectedSecureUnitButton.removeFunc = removeFunc;  
end


-- disable an action (if selected and close we open)
function ActionBar:DisableAction(actionPosition)
  -- check in combat
  local actionListSize = #self.actionList;
  if actionPosition<1 or actionPosition>actionListSize then
    error("intern error : actionPosition out of range");
    return;
  end
    
  self.state:DisableAction(actionPosition);
  self:SpreadState();  
end

-- simple add an action at the end (should evoluate to keep more information)
function ActionBar:EnableAction(actionPosition)
  -- check in combat
  local actionListSize = #self.actionList;
  if actionPosition<1 or actionPosition>actionListSize then
    error("intern error : actionPosition out of range");
    return;
  end
  
  self.state:EnableAction(actionPosition);
  self:SpreadState();
end



-- on click on this button will cast selected element
function ActionBar:GetSelectedButton()

end

-- returns {selectedAction=7, selectedVisibility=0, actionList={2, 1, 6, 7, 3, 0, 0}, visibilityList={2, 2, 2, 2, 2, 4, 4}}
-- this state can be given in constructor (actionBarState)
function ActionBar:GetBackup()
  return self.state;
end

function ActionBar:ACTION_AVAILABILITY_CHANGED(actionPosition, available)
  if available then
    self:EnableAction(actionPosition);
  else
    self:DisableAction(actionPosition);
  end
end

function ActionBar:SetActionBarRenderer(actionBarRenderer)

  -- UnApply old skin
  self.actionBarRenderer:UnApplySkin(self.selectedSecureUnitButton);
  for _, secureUnitButton in ipairs(self.secureUnitButtonList) do
    self.actionBarRenderer:UnApplySkin(secureUnitButton);  
  end

  -- Apply new skin
  self.actionBarRenderer = actionBarRenderer;  
  actionBarRenderer:ApplySkin(self.selectedSecureUnitButton);
  actionBarRenderer:ActionChanged(self.selectedSecureUnitButton);
  for _, secureUnitButton in ipairs(self.secureUnitButtonList) do
    actionBarRenderer:ApplySkin(secureUnitButton);
    actionBarRenderer:ActionChanged(secureUnitButton);
  end

end

--[[
  The ActionBarFactory class.
--]]

function ActionBarFactory:CreateSecureUnitButton(buttonName, parent, actionBar, position)
  local secureUnitButton = CreateFrame("Button", buttonName, parent, "SecureUnitButtonTemplate");
  secureUnitButton.actionBar = actionBar;
  secureUnitButton.borderTexture = secureUnitButton:CreateTexture(nil, "OVERLAY");
  
  secureUnitButton.targetText = secureUnitButton:CreateFontString(nil, "ARTWORK", "NumberFontNormal");
  secureUnitButton.targetText:SetAllPoints();
  
  secureUnitButton.cooldown = CreateFrame("Cooldown", nil, secureUnitButton, "CooldownFrameTemplate");  
    
  secureUnitButton.position = position;
  
  secureUnitButton:SetScript("OnEnter", function (button)
    local actionBarRenderer = button.actionBar.actionBarRenderer;
    if actionBarRenderer:IsTooltipAsked() then
      actionBarRenderer:UpdateTooltip(button);      
    end
  end);
  
  if position~=0 then
    secureUnitButton:SetScript("PreClick", function (button, click)
      local i = button.position;
      local actionBar = button.actionBar;
      local state = actionBar.state;
      local currentAction = state.actionList[i];
      local currentButton = button;
      if IsControlKeyDown() then
        local tmpButton = nil;
        local tmpAction;        
        if IsShiftKeyDown() then
          if i<state.totalCount then
            tmpButton = actionBar.secureUnitButtonList[i+1];
            tmpAction = state.actionList[i+1];
            state.actionList[i+1] = currentAction;
          end
        else
          if i>0 then
            tmpButton = actionBar.secureUnitButtonList[i-1];
            tmpAction = state.actionList[i-1];
            state.actionList[i-1] = currentAction;
          end
        end
        if tmpButton~=nil then
          state.actionList[i] = tmpAction;
          tmpButton.action = actionBar.actionList[currentAction];
          actionBar.actionBarRenderer:ActionChanged(tmpButton, actionBar.actionList[currentAction]);
          currentButton.action = actionBar.actionList[tmpAction];
          actionBar.actionBarRenderer:ActionChanged(currentButton);
        end
      elseif IsAltKeyDown() then
        -- nothing to do
      else
        if state:IsOpened() then
          actionBar.selectedSecureUnitButton.action = actionBar.actionList[currentAction];
          actionBar.actionBarRenderer:ActionChanged(actionBar.selectedSecureUnitButton);      
          state:Close(i);
        else
          if click=="RightButton" then
            state:Swap(i);
            actionBar.selectedSecureUnitButton.action = actionBar.actionList[currentAction];
            actionBar.actionBarRenderer:ActionChanged(actionBar.selectedSecureUnitButton);
            currentButton.action = actionBar.actionList[state.actionList[i]];
            actionBar.actionBarRenderer:ActionChanged(currentButton);
            actionBar.secureUnitButtonList[state.selectedPosition].action = actionBar.actionList[state.actionList[state.selectedPosition]];
            actionBar.actionBarRenderer:ActionChanged(actionBar.secureUnitButtonList[state.selectedPosition]);            
          end        
        end
      end
    end);

  
  else
  
    secureUnitButton:SetScript("PreClick", function (button, click)
      local actionBar = button.actionBar;
      local state = actionBar.state;
      if IsControlKeyDown() then
        -- remove action (no effet on state)
        if click=="LeftButton" and button.removeFunc~=nil then 
          button.removeFunc();
        end
      elseif IsAltKeyDown() then
        -- state add / remove favorite
        if IsShiftKeyDown() then
          state:RemoveFavorite();
        else
          state:AddFavorite();        
        end
      else
        if click=="RightButton" then
          state:Open();
        end
      end
    end)
    
  end

  secureUnitButton:SetScript("OnLeave", function ()
    GameTooltip:Hide();
  end);
  
  return secureUnitButton;
end

function ActionBarFactory:Create(parent, casterButtonName, actionList, actionBarRenderer, actionBarState)
  local actionBar = self:NewInstance();
  
  actionBar.actionList = actionList;
  
  local actionListSize = #actionList;
  
  if actionListSize == 0 then
    return nil;
  end


  -- this algo keep only available action but not all available action ()  
  if actionBarState==nil then
    actionBarState = actionBarStateFactory:Create(actionListSize);
  else
    actionBarStateFactory:Cast(actionBarState);
  end
  actionBar.state = actionBarState;
  
  for i=1, actionListSize do
    action = actionList[i];
    local actionAvailable = action:IsAvailable();
    local actionEnabled = actionBarState:IsActionEnabled(i);
    --     if actionEnabled ~= actionAvailable then BUG ??
    if actionEnabled and not actionAvailable then
      actionBarState:SetActionEnabled(i, actionAvailable);
    end
    if not actionEnabled and actionAvailable then
       actionBarState:SetActionEnabled(i, actionAvailable);
    end
  end
    
  local parent = CreateFrame("Button", nil, parent, "SecureHandlerClickTemplate");    
  actionBar.header = parent;
  
  local clickSnipset = [[
    local currentButton = self;
    if IsControlKeyDown() then
      local i=0;
      while buttons[i]~=self do
        i = i + 1;
      end      
      local tmpButton = nil;
      if IsShiftKeyDown() then
        if i<totalCount then
          tmpButton = buttons[i + 1];
        end
      else
        if i>0 then
          tmpButton = buttons[i - 1];
        end
      end
      if tmpButton~=nil then
        local tmpType = tmpButton:GetAttribute("type");
        local tmpSpell = tmpButton:GetAttribute("spell");
        local tmpUnit = tmpButton:GetAttribute("unit");
        tmpButton:SetAttribute("type", currentButton:GetAttribute("type"));
        tmpButton:SetAttribute("spell", currentButton:GetAttribute("spell"));
        tmpButton:SetAttribute("unit", currentButton:GetAttribute("unit"));
        currentButton:SetAttribute("type", tmpType);
        currentButton:SetAttribute("spell", tmpSpell);
        currentButton:SetAttribute("unit", tmpUnit);
      end
      return false;
    elseif IsAltKeyDown() then
        -- nothing with alt-key
        return false;
    else
      if button=="LeftButton" or button=="RightButton" then
        if open then
          local i=0;
          while buttons[i]~=self do
            i = i + 1;
          end      
          selectedPosition = i;
          for i=favoriteCount+1,totalCount do
            buttons[i]:Hide();
          end
          selectedButton:SetAttribute("type", currentButton:GetAttribute("type"));
          selectedButton:SetAttribute("spell", currentButton:GetAttribute("spell"));
          selectedButton:SetAttribute("unit", currentButton:GetAttribute("unit"));
          selectedButton:Show();
          open = false;
        else
          if button=="RightButton" then
            local fromButton = self;
            local toButton = selectedButton;
            toButton:SetAttribute("type", fromButton:GetAttribute("type"));
            toButton:SetAttribute("spell", fromButton:GetAttribute("spell"));
            toButton:SetAttribute("unit", fromButton:GetAttribute("unit"));
            fromButton = buttons[selectedPosition];
            toButton = self;
            toButton:SetAttribute("type", fromButton:GetAttribute("type"));
            toButton:SetAttribute("spell", fromButton:GetAttribute("spell"));
            toButton:SetAttribute("unit", fromButton:GetAttribute("unit"));
            fromButton = selectedButton;
            toButton = buttons[selectedPosition];
            toButton:SetAttribute("type", fromButton:GetAttribute("type"));
            toButton:SetAttribute("spell", fromButton:GetAttribute("spell"));
            toButton:SetAttribute("unit", fromButton:GetAttribute("unit"));
          end
        end
        -- do not cast if RightButton
        return button=="LeftButton";      
      else
        return false;      
      end
    end
  ]];

  local selectedClickSnipset = [[
    if IsControlKeyDown() then
      -- try to avoid action with a post click method
      return false;
    elseif IsAltKeyDown() then
      if IsShiftKeyDown() then
        -- remove favorite
        if favoriteCount>0 then
          buttons[favoriteCount]:Hide();
          favoriteCount = favoriteCount - 1;
        end
      else
        -- add favorite
        if favoriteCount<totalCount then
          favoriteCount = favoriteCount + 1;
          buttons[favoriteCount]:Show();
        end
      end
      return false;
    else
      if button=="LeftButton" then
        return true;      
      elseif button=="RightButton" then
        selectedPosition = 0;
        for i=favoriteCount+1,totalCount do
          buttons[i]:Show();
        end
        self:Hide();
        open = true;
        return false;
      else
        return false;      
      end
    end
  ]];


  local initSnipset = [[
    local allChilds = newtable(owner:GetChildren());
    local size = #allChilds - 1;
    selectedButton = allChilds[1];
    buttons = newtable();
    for i=1, size do
      buttons[i] = allChilds[i+1];    
    end
  ]];

     
  local selectedSecureUnitButton = self:CreateSecureUnitButton(casterButtonName, parent, actionBar, 0);
  actionBar.selectedSecureUnitButton = selectedSecureUnitButton;
  selectedSecureUnitButton:Hide();
  actionBarRenderer:ApplySkin(selectedSecureUnitButton);  
  selectedSecureUnitButton:RegisterForClicks("LeftButtonUp", "RightButtonUp");
  SecureHandlerWrapScript(selectedSecureUnitButton, "OnClick", parent, selectedClickSnipset);

  -- tmp variable to complete link between buttons
  local secureUnitButtonList = {}; -- visible buttons (created by renderer)
  actionBar.secureUnitButtonList = secureUnitButtonList;
  
  for i=1,actionListSize do
    -- keep actionListSize state (one by action)  
    local secureUnitButton = self:CreateSecureUnitButton("ActionBarLibSecureButton"..ActionBarFactory.lastSecureButtonId, parent, actionBar, i);
    ActionBarFactory.lastSecureButtonId = ActionBarFactory.lastSecureButtonId + 1;
    secureUnitButtonList[i] = secureUnitButton;
    actionBarRenderer:ApplySkin(secureUnitButton);
    secureUnitButton:RegisterForClicks("LeftButtonUp", "RightButtonUp");
    SecureHandlerWrapScript(secureUnitButton, "OnClick", parent, clickSnipset);
  end
      
  actionBar.actionBarRenderer = actionBarRenderer;
  
  parent:Execute(initSnipset);
      
  -- spread state
  actionBar:SpreadState();    
  
  for i=1,actionListSize do
    local action = actionList[i];
    -- can be directly enable, that why we register at the end
    action:RegisterEvent("ACTION_AVAILABILITY_CHANGED", actionBar, i);
  end
  
  return actionBar;  
end

ActionBarFactory.lastSecureButtonId = 1;

